import tkinter as tk
from functools import partial
from colorir import Palette, find_palettes
from colorir.utils import hue_sort_key

colors = Palette.load(palettes_dir=".")
palettes = {"all": colors}
palettes.update({
    pal_name: Palette.load(pal_name, palettes_dir=".")
    for pal_name in find_palettes(palettes_dir=".", kind=Palette)
})


class Window(tk.Tk):
    def __init__(self):
        super().__init__()
        self.geometry('1280x720')
        self.resizable(False, True)

        self.pal_btns = tk.Frame(self)
        self.pal_btns.pack(side=tk.LEFT, fill=tk.Y)
        for pal_name, palette in palettes.items():
            pal_name = pal_name.capitalize().replace('_', ' ') if pal_name.islower() else pal_name
            btn = tk.Button(self.pal_btns,
                            text=pal_name,
                            command=partial(self.add_color_btns, palette.color_names))
            btn.pack(fill='x')
        self.canvas = tk.Canvas(self)
        self.canvas.pack(side=tk.LEFT, expand=True, fill=tk.BOTH)
        scrollbar = tk.Scrollbar(self, orient=tk.VERTICAL, command=self.canvas.yview)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.canvas.configure(yscrollcommand=scrollbar.set)
        self.canvas.bind('<Configure>', lambda e: self.canvas.configure(
            scrollregion=self.canvas.bbox('all')
        ))
        frame = tk.Frame(self.canvas)
        self.canvas.create_window((0, 0), window=frame, anchor='nw')

        self.btn_size = 45
        # Create a button for each available color
        self.color_btns = {name: FramedButton(
            frame,
            width=self.btn_size,
            height=self.btn_size,
            bg=colors.get_color(name),  # Background of the button is its color
            bd=1,
            command=partial(self.btn_command, name)
        ) for name in colors.color_names}
        self.update()
        self.add_color_btns(colors.color_names)

    def btn_command(self, name):
        # Adds a strip of the selected color to the left of the screen
        self.pal_btns.config(bg=colors.get_color(name))
        self.clipboard_clear()
        self.clipboard_append(name)

    def add_color_btns(self, c_names):
        for btn in self.color_btns.values():
            btn.grid_forget()
        row = 0
        col = 0
        for c_name in sorted(c_names, key=lambda name: hue_sort_key(8)(colors.get_color(name))):
            self.color_btns[c_name].grid(row=row, column=col)
            if (col + 2) * self.btn_size < self.canvas.winfo_width():
                col += 1
            else:
                col = 0
                row += 1
        self.update_idletasks()
        self.canvas.configure(scrollregion=self.canvas.bbox('all'))


class FramedButton(tk.Frame):
    def __init__(self, parent, width, height, *args, **kwargs):
        super().__init__(parent, width=width, height=height)
        self.pack_propagate(0)
        self.button = tk.Button(self, *args, **kwargs)
        self.button.pack(fill=tk.BOTH, expand=1)


win = Window()
win.mainloop()



"""Configuration options.

Examples:
    Create a new color format and make it default:

    >>> from colorir import config, ColorFormat
    >>> config.DEFAULT_COLOR_FORMAT = ColorFormat(Hex, uppercase=True)

    Change the default locations from which palettes are loaded and to which they are saved:

    >>> config.DEFAULT_PALETTES_DIR = ".../my_project/palettes"
"""
from os import path

from .color_format import ColorFormat
from .color_class import Hex

DEFAULT_PALETTES_DIR = path.join(path.dirname(__file__), "palettes")
"""Default directory from which palettes will be loaded and to which they will be saved."""

DEFAULT_COLOR_FORMAT = ColorFormat(Hex)
"""Default color format used by different objects in this package."""


"""Palettes to hold collections of colors that can be shared between projects.

The :class:`Palette` class provides an easy way to manage your favorite colors in different
projects. In this context, a palette should be understood as any collection of colors that
can be grouped due to a common feature, not only colors that necessarily "look good" together.

Every color in a :class:`Palette` has a name associated with it. If unnamed colors are better fit
to your particular use case, you may want to use a :class:`StackPalette` instead.

Examples:
    Create a palette with the color red:

    >>> palette = Palette(red="#ff0000")

    Add the color blue:

    >>> palette.add("blue", "#0000ff")

    Get the value for red:

    >>> palette.red
    Hex('#ff0000')

    Get names of all colors:

    >>> palette.color_names
    ['red', 'blue']

    Remove red:

    >>> palette.remove("red")
    >>> "red" in palette.color_names
    False

    Name the palette and save it to the default directory:

    >>> palette.name = "single_blue"
    >>> palette.save()

    Load it again from elsewhere later:

    >>> palette = Palette.load("single_blue")
"""
import abc
import json
import os
from pathlib import Path
from typing import Union, List
from warnings import warn

from . import config
from . import utils
from .color_class import ColorBase, sRGB, HSL, HSV
from .color_format import ColorFormat, ColorLike

_throw_exception = object()
_builtin_palettes_dir = Path(__file__).resolve().parent / "builtin_palettes"

__all__ = [
    "Palette",
    "StackPalette",
    "find_palettes",
    "delete_palette"
]


class PaletteBase(metaclass=abc.ABCMeta):
    def __init__(self, name=None, color_format=None):
        if color_format is None:
            color_format = config.DEFAULT_COLOR_FORMAT

        self.name = name
        self._color_format = color_format

    @property
    @abc.abstractmethod
    def colors(self):
        """A list of all color values currently stored in the palette."""
        pass

    @property
    def color_format(self) -> ColorFormat:
        """Color format specifying how the colors of this palette are stored."""
        return self._color_format

    @color_format.setter
    @abc.abstractmethod
    def color_format(self, color_format):
        self._color_format = color_format

    def __len__(self):
        return len(self.colors)

    def __contains__(self, item):
        return self.color_format.format(item) in self.colors

    def __iter__(self):
        return iter(self.colors)

    def most_similar(self, color: ColorLike, n=1, method="CIE76"):
        """Finds the `n` most similar colors to `color` in this palette.

        Args:
            color: The value of the color of reference. Can be an instance of any
                :mod:`~colorir.color_class` class or, alternatively, a color-like object that
                resembles the color to which others will be compared in search of similar results.
            n: How many similar colors to be retrieved. -1 means all colors from the palette will
                be returned from most similar to least.
            method: Method for calculating color distance. See the documentation of
                :func:`~colorir.utils.color_dist()`.

        Examples:
            >>> palette = Palette(red="#ff0000", blue="#0000ff")
            >>> palette.most_similar("#880000")
            Hex('#ff0000')

        Returns:
            A single :class:`~colorir.color_class.ColorBase` if `n` == 1 or a list of
            :class:`~colorir.color_class.ColorBase` if `n` != 1. If the return type is a list, the
            colors will be ordered from most similar to least.
        """
        color = self.color_format.format(color)
        closest = sorted(self.colors,
                         key=lambda color2: utils.color_dist(color, color2, method=method))
        if n == 1:
            return closest[0]
        elif n < 0:
            n = len(self)
        return closest[:n]

    def to_cmap(self, N: int = None):
        """Converts this palette into a matplotlib ListedColormap.

        Args:
            N: Passed down to ListedColormap constructor.
        """
        from matplotlib.colors import ListedColormap

        colors = [color.hex(include_a=True, tail_a=True) for color in self.colors]
        if self.name is None:
            return ListedColormap(colors, N=N)
        return ListedColormap(colors, N=N, name=self.name)


class Palette(PaletteBase):
    """Class that holds colors values associated with names.

    Examples:
        >>> palette = Palette(red="#ff0000") # Uses default color format
        >>> palette.red
        Hex('#ff0000')

        For more examples see the documentation of the :mod:`~colorir.palette` module.

    Args:
        name: Name of the palette which will be used to save it with :meth:`Palette.save()`.
        color_format: Color format specifying how the colors of this :class:`Palette` should be
            stored. Defaults to the value specified in
            :const:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        colors: Colors that will be stored in this palette.
    """
    def __init__(self,
                 name: str = None,
                 color_format: ColorFormat = None,
                 **colors: ColorLike):
        super().__init__(name, color_format)

        self._color_dict = {}
        for k, v in colors.items():
            self.add(k, v)

    @property
    def colors(self) -> List[ColorBase]:
        return list(self._color_dict.values())

    @property
    def color_names(self) -> List[str]:
        """A list of all color names currently stored in the palette."""
        return list(self._color_dict.keys())

    # color_format could be used to build a color on every Palette.color call, but that is
    # computationally intensive. That's why the colors are stored as ready objects and are
    # re-created if needed
    @PaletteBase.color_format.setter
    def color_format(self, color_format):
        PaletteBase.color_format.fset(self, color_format)

        self._color_dict = {c_name: color_format._from_rgba(c_value._rgba)
                            for c_name, c_value in self._color_dict.items()}

    @classmethod
    def load(cls,
             palettes: Union[str, List[str]] = None,
             palettes_dir: str = None,
             search_builtins=True,
             name: str = None,
             color_format: ColorFormat = None,
             warnings=True):
        """Factory method that loads previously created palettes into a :class:`Palette` instance.

        A palette is a file containing json-formatted information about colors that ends with the
        '.palette' extension. You should not create such files manually but rather through the
        :meth:`Palette.save()` method.

        If multiple palettes define different color values under the same name, only the first one
        will be kept. You can define the order in which the palettes are loaded by ordering them in
        the `palettes` parameter. By default this occurrence logs a warning, but this behaviour can
        be changed through the `warnings` parameter.

        Args:
            palettes: List of palettes located in the location represented by `palettes_dir` that
                should be loaded by this :class:`Palette` instance. Additionally may include
                built-in palettes such as 'css' if `search_builtins` is set to ``True``. If this
                parameter is a string, the :attr:`Palette.name` will be inferred from it. By
                default, loads all palettes found in the specified directory.
            palettes_dir: The directory from which the palettes specified in the `palettes`
                parameter will be loaded. Defaults to the value specified in
                :data:`config.DEFAULT_PALETTES_DIR <colorir.config.DEFAULT_PALETTES_DIR>`.
            search_builtins: Whether `palettes` also includes built-in palettes such as 'css' or
                'basic'. Set to ``False`` to ensure only palette files found in `palettes_dir` are
                loaded.
            name: Name of the palette which will be used to save it with the
                :meth:`Palette.save()`. If the `palettes` parameter is a single string, defaults
                to that.
            color_format: Color format specifying how the colors of this :class:`Palette` should be
                stored. Defaults to the value specified in
                :data:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
            warnings: Whether to emit a warning if two colors with the same name but different
                color values were found in the `palettes`.

        Examples:
            Loads the default CSS palette:

            >>> css_palette = Palette.load('css')

            Loads both the basic and fluorescent palettes into a new palette called 'colorful':

            >>> colorful = Palette.load(['basic', 'fluorescent'], name='colorful')
        """
        if palettes_dir is None:
            palettes_dir = config.DEFAULT_PALETTES_DIR
        palettes_dir = [palettes_dir]
        if search_builtins:
            palettes_dir.append(_builtin_palettes_dir)
        if isinstance(palettes, str):
            if name is None:
                name = palettes
            palettes = [palettes]

        found_palettes = {}
        for path in palettes_dir:
            path = Path(path)
            palette_files = path.glob("*.palette")
            for palette_file in palette_files:
                palette_name = palette_file.name.replace(".palette", '')
                found_palettes[palette_name] = json.loads(palette_file.read_text())

        palette_obj = cls(name=name, color_format=color_format)
        if palettes is None: palettes = list(found_palettes)
        # Reiterates based on user input order
        for palette_name in palettes:
            for c_name, c_rgba in found_palettes[palette_name].items():
                c_rgba = (int(c_rgba[3:5], 16),
                          int(c_rgba[5:7], 16),
                          int(c_rgba[7:9], 16),
                          int(c_rgba[1:3], 16))
                new_color = palette_obj.color_format._from_rgba(c_rgba)
                old_color = palette_obj.get_color(c_name, new_color)
                if new_color != old_color and warnings:
                    warn(
                        f"a discrepancy was detected when adding color '{c_name}' "
                        f"({new_color}) from palette named '{palette_name}': '{c_name}' is "
                        f"already present with a different value ({old_color})")
                else:
                    palette_obj.add(c_name, new_color)
        return palette_obj

    def __getattr__(self, item):
        return self._color_dict[item]

    def __repr__(self):
        name_str = self.name + ", " if self.name else ""
        color_strs = [f"{c_name}={c_val.__repr__()}" for c_name, c_val in self._color_dict.items()]
        return f"{self.__class__.__name__}({name_str}{', '.join(color_strs)})"

    def __eq__(self, other):
        return self._color_dict.items() == other._color_dict.items()

    def __add__(self, other):
        for c_name in other.color_names:
            self.add(c_name, other.get_color(c_name))
        return self

    def get_color(self,
                  name: Union[str, List[str]],
                  fallback=_throw_exception) -> Union[ColorBase, List[ColorBase]]:
        """Retrieves one or more colors from the :class:`Palette` given their names.

        Args:
            name: One or more names of colors in this palette.
            fallback: What to return instead of colors that are not found in the palette. By
                default, throws an exception.

        Examples:
            >>> palette = Palette(red="#ff0000", blue="#0000ff")
            >>> palette.get_color("red")
            Hex('#ff0000')
            >>> palette.get_color(["red", "blue"])
            [Hex('#ff0000'), Hex('#0000ff')]

        Returns:
            A single :class:`~colorir.color_class.ColorBase` if `name` is a string or a list of
            :class:`~colorir.color_class.ColorBase` if `name` is a list of strings.
        """
        if fallback is _throw_exception:
            if isinstance(name, str):
                return self._color_dict[name]
            return [self._color_dict[name_i] for name_i in name]
        if isinstance(name, str):
            return self._color_dict.get(name, fallback)
        return [self._color_dict.get(name_i, fallback) for name_i in name]

    def get_names(self, color: ColorLike) -> list:
        """Finds all names of the provided color in this :class:`Palette`.

        Compares the provided `color` to every color the :class:`Palette` contains and returns the
        names of the colors that are equivalent to the one provided.

        Args:
            color: The value of the color to be searched for. Can be an instance of any
                :mod:`~colorir.color_class` class or, alternatively, a color-like object that
                resembles the color you want to search for.

        Examples:
            >>> palette = Palette(red="#ff0000")
            >>> palette.get_names(HSL(0, 1, 0.5))
            ['red']
        """
        color = self.color_format.format(color)
        color_list = []
        for name in self.color_names:
            if self.get_color(name) == color:
                color_list.append(name)
        return color_list

    def add(self, name: str, color: ColorLike):
        """Adds a color to the palette.

        Two colors with the same name but different values are invalid and can not coexist in a
        same :class:`Palette`. You should therefore avoid reusing names of already existing
        colors.

        Args:
            name: Name to be assigned to the new color.
            color: The value of the color to be created. Can be an instance of any
                :mod:`~colorir.color_class` class or, alternatively, a color-like object that
                resembles the color you want to add.

        Examples:
            Adding a new blue color to the palette:

            >>> palette = Palette()
            >>> palette.add("bestblue", "4287f5")
            >>> palette.bestblue
            Hex('#4287f5')
        """
        # Test to detect invalid color names
        if name in dir(self):
            raise ValueError(f"'{name}' is a reserved attribute name of Palette or is already "
                             "in use by another color. 'Palette.update()' might help in that case")
        color = self.color_format.format(color)
        self._color_dict[name] = color

    def update(self, name: str, color: ColorLike):
        """Updates a color to a new value.

        Args:
            name: Name of the color to be updated.
            color: The value of the color to be updated. Can be an instance of any
                :mod:`~colorir.color_class` class or, alternatively, a color-like object that
                resembles the format of the color you want to update.

        Examples:
            Create a slightly dark shade of red:
            
            >>> palette = Palette(myred="dd0000")
            >>> palette.myred
            Hex('#dd0000')
            
            Change it to be even a bit darker:
            
            >>> palette.update("myred", "800000")
            >>> palette.myred
            Hex('#800000')
        """
        if name in self._color_dict:
            self._color_dict[name] = self.color_format.format(color)
        else:
            raise ValueError(f"provided 'name' parameter is not a color loaded in this 'Palette'")

    def remove(self, name):
        """Removes a color from the palette.

        Args:
            name: Name of the color to be removed.

        Examples:
            >>> palette = Palette(red=sRGB(255, 0, 0))
            >>> palette.remove("red")
            >>> "red" in palette.color_names
            False
        """
        if name in self._color_dict:
            del self._color_dict[name]
        else:
            raise ValueError(f"provided 'name' parameter is not a color stored in this 'Palette'")

    def save(self, palettes_dir: str = None):
        """Saves the changes made to this :class:`Palette` instance.

        If this method is not called after modifications made by :meth:`Palette.add()`,
        :meth:`Palette.update()` and :meth:`Palette.remove()`, the modifications on the palette
        will not be permanent.

        Examples:
            Loads both the basic and fluorescent palettes into a new palette called 'colorful':

            >>> colorful = Palette.load(['basic', 'fluorescent'], name='colorful')

            Save the palette to the default palette directory:

            >>> colorful.save()
        """
        if self.name is None:
            raise AttributeError(
                "the 'name' attribute of a 'Palette' instance must be defined to save it")
        if palettes_dir is None:
            palettes_dir = config.DEFAULT_PALETTES_DIR
        with open(Path(palettes_dir) / (self.name + ".palette"), "w") as file:
            formatted_colors = {}
            for c_name, c_val in self._color_dict.items():
                c_rgba = c_val.hex(include_a=True, tail_a=False)
                formatted_colors[c_name] = c_rgba
            json.dump(formatted_colors, file, indent=4)

    def to_stackpalette(self) -> "StackPalette":
        """Converts this palette into a :class:`StackPalette`."""
        return StackPalette(self.name, self.color_format, *self._color_dict.values())

    def to_dict(self):
        """Converts this palette into a python `dict`."""
        return dict(self._color_dict)


class StackPalette(PaletteBase):
    """Class that handles anonymous indexed colors stored as a stack.

    This class may be used as a replacement for :class:`Palette` when the name of the colors is
    irrelevant.

    Examples:
        >>> spalette = StackPalette("elementary", None, "ff0000", "00ff00", "0000ff")
        >>> spalette[0]
        Hex('#ff0000')

    Args:
        name: Name of the palette which will be used to save it with
            :meth:`StackPalette.save()`.
        color_format: Color format specifying how the colors of this :class:`StackPalette` should
            be stored. Defaults to the value specified in
            :data:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        colors: Colors that will be stored in this palette.
    """

    def __init__(self,
                 name: str = None,
                 color_format: ColorFormat = None,
                 *colors: ColorLike):
        super().__init__(name=name, color_format=color_format)

        self._color_stack = []
        for color in colors:
            self.add(color)

    @property
    def colors(self) -> List[ColorBase]:
        """colors: A list of all color values currently stored in the :class:`StackPalette`."""
        return list(self._color_stack)

    @PaletteBase.color_format.setter
    def color_format(self, color_format):
        PaletteBase.color_format.fset(self, color_format)

        self._color_stack = [color_format._from_rgba(color._rgba) for color in self._color_stack]

    @classmethod
    def load(cls,
             palettes: Union[str, List[str]] = None,
             palettes_dir: str = None,
             search_builtins=True,
             name: str = None,
             color_format: ColorFormat = None):
        """Factory method that loads previously created stack palettes into a
        :class:`StackPalette` instance.

        A stack palette is a file containing json-formatted information about colors that ends
        with the '.spalette' extension. You should not create such files manually but rather
        through the :meth:`StackPalette.save()` method.

        Examples:
            Load a stack palette called "project_interface" from the default directory:

            >>> spalette = StackPalette.load("project_interface")  # doctest: +SKIP

        Args:
            palettes: List of stack palettes located in the location represented by `palettes_dir`
                that should be loaded by this :class:`StackPalette` instance. If this parameter is
                a string, the :attr:`StackPalette.name` will be inferred from it. By default,
                loads all palettes found in the specified directory.
            palettes_dir: The directory from which the palettes specified in the `palettes`
                parameter will be loaded. Defaults to the value specified in
                :data:`config.DEFAULT_PALETTES_DIR <colorir.config.DEFAULT_PALETTES_DIR>`.
            search_builtins: Whether `palettes` also includes built-in palettes such as 'tab10' or
                'dark2'. Set to ``False`` to ensure only palette files found in `palettes_dir` are
                loaded.
            name: Name of the palette which will be used to save it with the
                :meth:`StackPalette.save()`. If the `palettes` parameter is a single string,
                defaults to that.
            color_format: Color format specifying how the colors of this :class:`Palette` should be
                stored. Defaults to the value specified in
                :data:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        """
        if palettes_dir is None:
            palettes_dir = config.DEFAULT_PALETTES_DIR
        palettes_dir = [palettes_dir]
        if search_builtins:
            palettes_dir.append(_builtin_palettes_dir)
        if isinstance(palettes, str):
            if name is None:
                name = palettes
            palettes = [palettes]

        found_palettes = {}
        for path in palettes_dir:
            path = Path(path)
            palette_files = path.glob("*.spalette")
            for palette_file in palette_files:
                palette_name = palette_file.name.replace(".spalette", '')
                found_palettes[palette_name] = json.loads(palette_file.read_text())

        palette_obj = cls(name=name, color_format=color_format)
        if palettes is None: palettes = list(found_palettes)
        # Reiterates based on user input order
        for palette_name in palettes:
            for c_rgba in found_palettes[palette_name]:
                c_rgba = (int(c_rgba[3:5], 16),
                          int(c_rgba[5:7], 16),
                          int(c_rgba[7:9], 16),
                          int(c_rgba[1:3], 16))
                new_color = palette_obj.color_format._from_rgba(c_rgba)
                palette_obj.add(new_color)
        return palette_obj

    # TODO enhance to allow different color systems and variance param (prob. 1.4.0)
    @classmethod
    def new_complementary(cls,
                          n: int,
                          color: ColorLike = None,
                          name: str = None,
                          color_format: ColorFormat = None):
        """Creates a new palette with `n` complementary colors.

        Colors are considered complementary if they are interspaced in the additive HUE color
        wheel.

        Examples:
             Make a palette from red and its complementary color, cyan:

             >>> spalette = StackPalette.new_complementary(2, sRGB(255, 0, 0))
             >>> spalette
             StackPalette(Hex('#ff0000'), Hex('#00ffff'))

             Make a tetradic palette of random colors:

             >>> spalette = StackPalette.new_complementary(4)

        Args:
            n: The number of colors in the new palette.
            color: A color from which the others will be generated against. By default, a color is
                randomly chosen.
            name: Name of the palette which will be used to save it with
                :meth:`StackPalette.save()`.
            color_format: Color format specifying how the colors of this :class:`StackPalette`
                should be stored. Defaults to the value specified in
                :data:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        """
        n_spalette = cls(name=name, color_format=color_format)
        if color is None:
            hsv = utils.random_color(color_format=ColorFormat(HSV, max_h=360))
        else:
            hsv = n_spalette.color_format.format(color).hsv(max_h=360)

        step = 360 / n
        for i in range(n):
            hue = (hsv[0] + i * step) % 360
            n_spalette.add(HSV(hue, hsv[1], hsv[2]))
        return n_spalette

    @classmethod
    def new_analogous(cls,
                      n: int,
                      sections=12,
                      start=0,
                      color: ColorLike = None,
                      name: str = None,
                      color_format: ColorFormat = None):
        """Creates a new palette with `n` analogous colors.

        Colors are considered analogous if they are side-by-side in the additive HUE color wheel.

        Examples:
             Make a palette from red and its analogous color, orange:

             >>> spalette = StackPalette.new_analogous(2, start=1, color=sRGB(255, 0, 0))
             >>> spalette
             StackPalette(Hex('#ff0000'), Hex('#ff8000'))

             Make a palette of four similar colors:

             >>> spalette = StackPalette.new_analogous(4, sections=24)

        Args:
            n: The number of colors in the new palette.
            sections: The number of sections in which the additive HUE color wheel will be divided
                before sampling colors. The bigger this number, the more similar the colors will
                be.
            start: Where the color described in the 'color' parameter will be placed with respect
                to the others. If '0', 'color' will be in the center of the generated palette, and
                colors will be sampled from both its sides in the HUE wheel. If '1', colors will
                be sampled clockwise from 'color'. If '-1', they will be sampled counter-clockwise.
            color: A color from which the others will be generated against. By default, a color is
                randomly chosen.
            name: Name of the palette which will be used to save it with
                :meth:`StackPalette.save()`.
            color_format: Color format specifying how the colors of this :class:`StackPalette`
                should be stored. Defaults to the value specified in
                :data:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.
        """
        if n > sections:
            raise ValueError("'n' parameter cannot be larger than 'sections' parameter")
        if start == 0:
            first = -int(n / 2)
            iterator = range(first, first + n)
        elif start == 1:
            iterator = range(n)
        elif start == -1:
            iterator = range(-n + 1, 1)
        else:
            raise ValueError("'starting_point' must be either 0, 1 or -1")

        n_spalette = cls(name=name, color_format=color_format)
        if color is None:
            hsv = utils.random_color(color_format=ColorFormat(HSV, max_h=360))
        else:
            hsv = n_spalette.color_format.format(color).hsv(max_h=360)

        step = 360 / sections
        for index, i in enumerate(iterator):
            hue = (hsv[0] + i * step) % 360
            n_spalette.add(HSV(hue, hsv[1], hsv[2]))
        return n_spalette

    def __getitem__(self, item: int):
        return self._color_stack[item]

    def __repr__(self):
        name_str = self.name + ", " if self.name else ""
        return f"{self.__class__.__name__}({name_str}" \
               f"{', '.join(c_val.__repr__() for c_val in self._color_stack)})"

    def __eq__(self, other):
        return self._color_stack == other._color_stack

    def __add__(self, other):
        for color in other:
            self.add(color)
        return self

    def swap(self, index1: int, index2: int):
        """Swap the places of two colors in the palette.

        Can be used to reorganize the palette if needed.

        Examples:
            >>> spalette = StackPalette(None, None, "ff0000", "0000ff")
            >>> spalette
            StackPalette(Hex('#ff0000'), Hex('#0000ff'))
            >>> spalette.swap(0, 1)
            >>> spalette
            StackPalette(Hex('#0000ff'), Hex('#ff0000'))

        Args:
            index1: The index of the first color.
            index2: The index of the second color.
        """
        c_temp = self._color_stack[index1]
        self._color_stack[index1] = self._color_stack[index2]
        self._color_stack[index2] = c_temp

    def add(self, color: ColorLike):
        """Adds a color to the end of the stack palette.

        Colors can only be added to the last index of the stack palette, just like in a normal
        stack.

        Args:
            color: The value of the color to be created. Can be an instance of any
                :mod:`~colorir.color_class` class or, alternatively, a color-like object that
                resembles the color you want to add.

        Examples:
            Adding a new blue color to the palette:

            >>> spalette = StackPalette()
            >>> spalette.add("4287f5")
            >>> spalette[0]
            Hex('#4287f5')
        """
        self._color_stack.append(self.color_format.format(color))

    def update(self, index: int, color: ColorLike):
        """Updates a color to a new value.

        Args:
            index: Index of the color to be updated.
            color: The value of the color to be updated. Can be an instance of any
                :mod:`~colorir.color_class` class or, alternatively, a color-like object that
                resembles the format of the color you want to update.

        Examples:
            Create a slightly dark shade of red:

            >>> spalette = StackPalette(None, None, "dd0000")
            >>> spalette[0]
            Hex('#dd0000')

            Change it to be even a bit darker:

            >>> spalette.update(0, "800000")
            >>> spalette[0]
            Hex('#800000')
        """
        self._color_stack[index] = self.color_format.format(color)

    def remove(self):
        """Removes a color from the end of the palette.

        Colors can only be removed from the last index of the stack palette, just like in a normal
        stack.

        Examples:
            >>> spalette = StackPalette(None, None, "#ff0000")
            >>> "#ff0000" in spalette
            True
            >>> spalette.remove()
            >>> "#ff0000" in spalette
            False
        """
        self._color_stack.pop()

    def save(self, palettes_dir: str = None):
        """Saves the changes made to this :class:`StackPalette` instance.

        If this method is not called after modifications made by :meth:`StackPalette.add()`,
        :meth:`StackPalette.update()` and :meth:`StackPalette.remove()`, the modifications on the
        palette will not be permanent.

        Examples:
            Create a new :class:`StackPalette` and save it to the current directory:

            >>> spalette = StackPalette("elementary", None, "ff0000", "00ff00", "0000ff")
            >>> spalette.save()
        """
        if self.name is None:
            raise AttributeError(
                "the 'name' attribute of a 'StackPalette' instance must be defined to save it"
            )
        if palettes_dir is None:
            palettes_dir = config.DEFAULT_PALETTES_DIR
        with open(Path(palettes_dir) / (self.name + ".spalette"), "w") as file:
            formatted_colors = []
            for c_val in self._color_stack:
                c_rgba = tuple(spec for spec in c_val._rgba)
                c_rgba = "#%02x" % c_rgba[-1] + "%02x%02x%02x" % c_rgba[:3]
                formatted_colors.append(c_rgba)
            json.dump(formatted_colors, file, indent=4)

    def to_palette(self, names: List[str]) -> Palette:
        """Converts this stack palette into a :class:`Palette`.

        Args:
            names: Names that will be given to the colors in the same order they appear in this
                stack palette.
        """
        if len(names) == len(set(names)) == len(self._color_stack):
            return Palette(self.name, self.color_format, **dict(zip(names, self._color_stack)))
        raise ValueError("'names' must have the same length as this 'StackPalette' and no "
                         "duplicates")


def find_palettes(palettes_dir: str = None,
                  search_builtins=True,
                  kind=(Palette, StackPalette)) -> List[str]:
    """Returns the names of the palettes found in `directory`.

    Args:
        palettes_dir: The directory from which the palettes will be searched for. Defaults to the
            value specified in
            :data:`config.DEFAULT_PALETTES_DIR <colorir.config.DEFAULT_PALETTES_DIR>`.
        search_builtins: Whether to also include built-in palettes such as 'css' or
            'basic' in the search.
        kind: The kinds of palettes to include in the search. Can be either :class:`Palette`,
            :class:`StackPalette`, or a list of any of those.
    """
    if palettes_dir is None:
        palettes_dir = config.DEFAULT_PALETTES_DIR
    if isinstance(palettes_dir, str):
        palettes_dir = [palettes_dir]
    if search_builtins:
        palettes_dir.append(_builtin_palettes_dir)
    globs = []
    if not isinstance(kind, (tuple, list)):
        kind = [kind]
    if Palette in kind:
        globs.append("*.palette")
    if StackPalette in kind:
        globs.append("*.spalette")

    palettes = []
    for path in palettes_dir:
        path = Path(path)
        for glob in globs:
            for file in path.glob(glob):
                palettes.append(file.name[:file.name.index('.')])
    return palettes


def delete_palette(palette: str, palettes_dir: str = None):
    """Permanently deletes the file associated with a palette.

    Be careful when using this function, there is no way to recover a palette after it has been
    deleted.

    Args:
        palette: The name of the palette that will be deleted.
        palettes_dir: The directory containing the palette to be deleted.
            Defaults to the value specified in
            :data:`config.DEFAULT_PALETTES_DIR <colorir.config.DEFAULT_PALETTES_DIR>`.
    """
    if palettes_dir is None:
        palettes_dir = config.DEFAULT_PALETTES_DIR

    path = Path(palettes_dir)
    palettes = list(path.glob(f"{palette}.*"))
    if len(palettes) == 1:
        os.remove(path / palettes[0])
    elif len(palettes) == 0:
        raise ValueError(f"couldn't find palette '{palette}' in '{palettes_dir}'")
    else:
        raise ValueError(f"palette name '{palette}' is ambiguous (more than one palette share it)")


import os
from math import sqrt
from random import randint
from typing import Union, List
from colormath.color_conversions import convert_color
from colormath.color_diff import *
from colormath.color_objects import sRGBColor, LabColor

from . import config
from . import palette
from .color_class import ColorBase, HCLab
from .color_format import ColorLike, ColorFormat
from .gradient import Grad

__all__ = [
    "swatch",
    "simplified_dist",
    "color_dist",
    "random_color",
    "color_str",
    "hue_sort_key"
]


def swatch(obj: Union[ColorLike, List[ColorLike], "palette.Palette", "palette.StackPalette"],
           colored_text=True,
           width=3,
           height=1,
           tabular=True):
    """Prints swatches of `obj` in the terminal with colored text.

    Each swatch consists of a rectangle of a color followed by its value (and name if known).

    Args:
        obj: What will be represented in the terminal. Is either a single color, a list of colors,
            a :class:`~colorir.palette.Palette`, a :class:`~colorir.palette.StackPalette` or a
            subclass of :class:`~colorir.gradient.Grad`.
        colored_text: Whether the text that follows the colored rectangles should also be colored.
        width: The width (in space characters) of the colored rectangles.
        height: The height (in number of lines) of the colored rectangles.
        tabular: Whether the colored rectangle, color name and color value should be printed each
            in its separate column. Only used if `obj` is a :class:`~colorir.palette.Palette`.
    """
    # Assume single ColorLike
    if isinstance(obj, (ColorBase, str, tuple)):
        obj = [obj]
    elif isinstance(obj, Grad):
        # Seven steps for each color transition = 7 * (n - 1) - (n - 2)
        obj = obj.n_colors(6 * len(obj.colors) - 5, include_ends=True)
    elif isinstance(obj, palette.Palette):
        longest_name = max([len(name) for name in obj.color_names])
    # Needed to make Windows understand "\33" (https://stackoverflow.com/questions/12492810/python-
    # how-can-i-make-the-ansi-escape-codes-to-work-also-in-windows)
    os.system("")
    for i, c_val in enumerate(obj):
        rect_str = color_str(" " * width, bg_color=c_val)
        val_str = f" {c_val}"
        if isinstance(obj, palette.Palette):
            name = obj.color_names[i]
            spacing = tabular * " " * (longest_name - len(name))
            val_str = ' ' + name + spacing + val_str
        if colored_text:
            val_str = color_str(val_str, fg_color=c_val)
        print(rect_str + val_str)
        for _ in range(height - 1):
            print(rect_str)


# Implemented this way rather than a sort_colors function because it can be combined with
# other keys (see color_picker example)
def hue_sort_key(hue_classes=1,
                 gray_thresh=12.0,
                 gray_start=True,
                 alt_lum=False,
                 invert_lum=False):
    """Returns a function that can be used as a key for python's 'sort' and 'sorted' in order to
    sort color-like objects by their hue component.

    Examples:
        >>> sorted(["0000ff", "ff0000",  "000000", "00ff00", "ffffff"], key=hue_sort_key())
        ['000000', 'ffffff', 'ff0000', '00ff00', '0000ff']

    Args:
        hue_classes: Number hue categories. Inside each hue category colors will be sorted by
            luminance rather than hue.
        gray_thresh: Chroma threshold bellow which a color will be considered a shade of gray.
        gray_start: Whether the colors considered shades of gray will be grouped at the start or
            end of the sorted iterable.
        alt_lum: Whether to alternate luminance values with each hue class transition. Only has an
            effect if `hue_classes` > 1.
        invert_lum: By default, sorting within hue classes happen from darker to lighter colors.
            This parameter allows inverting this patern, thus starting with light colors and going
            towards darker tones. Only has an effect if `hue_classes` > 1.

    Notes:
        The refered hue and chroma components are those of :class:`~colorir.color_class.HCLab`.
    """
    color_format = config.DEFAULT_COLOR_FORMAT
    gray_hue = -1 if gray_start else hue_classes + 1

    def sort_key(color):
        interpreted = color_format.format(color)
        h, c, l = HCLab._from_rgba(interpreted._rgba, max_h=1)
        if c < gray_thresh:
            h = gray_hue
            if not gray_start and alt_lum and hue_classes % 2 == 1:
                l *= -1
        elif hue_classes > 1:
            h = int(h * hue_classes)
            if alt_lum and h % 2 == (not gray_start):
                l *= -1

        if hue_classes == 1:
            return h
        if invert_lum:
            l *= -1
        return h, l

    return sort_key


def simplified_dist(color1: ColorLike,
                    color2: ColorLike):
    """Calculates the perceived distance between two colors.

    There are many methods to approach the similarity of colors mathematically. The
    algorithm implemented in this function [#]_ tries to provide balance between efficiency and
    accuracy by using a weighted euclidian distance formula in the RGB color space.

    References:
        .. [#] Adapted from "Colour metric" by Thiadmer Riemersma. Available on
            https://www.compuphase.com/cmetric.htm.

    Args:
        color1: First color point.
        color2: Second color point.
    """
    color_format = config.DEFAULT_COLOR_FORMAT
    # We only need the '._rgba's, so no need to convert if already ColorBase
    if not isinstance(color1, ColorBase):
        color1 = color_format.format(color1)
    if not isinstance(color2, ColorBase):
        color2 = color_format.format(color2)
    rgba1, rgba2 = color1._rgba, color2._rgba
    avg_r = (rgba1[0] + rgba2[0]) / (2 * 255)
    d_r = rgba1[0] - rgba2[0]
    d_g = rgba1[1] - rgba2[1]
    d_b = rgba1[2] - rgba2[2]
    return sqrt((2 + avg_r) * d_r ** 2
                 + 4 * d_g ** 2
                 + (3 - avg_r) * d_b ** 2)


# TODO doc (mention 2000 not working properly in colormath and kwargs for delta-e funcs)
def color_dist(color1: ColorLike, color2: ColorLike, method="CIE76"):
    if method == "simplified":
        return simplified_dist(color1, color2)

    color_format = config.DEFAULT_COLOR_FORMAT
    if not isinstance(color1, ColorBase):
        color1 = color_format.format(color1)
    if not isinstance(color2, ColorBase):
        color2 = color_format.format(color2)
    color1 = sRGBColor(*color1._rgba[:3], is_upscaled=True)
    color2 = sRGBColor(*color2._rgba[:3], is_upscaled=True)
    color1 = convert_color(color1, LabColor, target_illuminant="d65")
    color2 = convert_color(color2, LabColor, target_illuminant="d65")

    if method == "CIE76":
        return delta_e_cie1976(color1, color2)
    if method == "CIE94":
        return delta_e_cie1994(color1, color2)
    if method == "CIE2000":
        return delta_e_cie2000(color1, color2)
    if method == "CMC":
        return delta_e_cmc(color1, color2)

    raise ValueError("invalid 'method' parameter")


def random_color(random_a=False,
                 color_format: ColorFormat = None):
    """Generates a new random color.

    Args:
        random_a: Whether to randomize the alpha attribute as well or just make it 1.
        color_format: Specifies the format of the output color. Defaults to
            :data:`config.DEFAULT_COLOR_FORMAT <colorir.config.DEFAULT_COLOR_FORMAT>`.

    Examples:
        >>> random_color()  # doctest: +SKIP
        Hex('#304fcc')
    """
    if color_format is None:
        color_format = config.DEFAULT_COLOR_FORMAT
    if random_a:
        a = randint(0, 255)
    else:
        a = 255
    return color_format._from_rgba((randint(0, 255),
                                    randint(0, 255),
                                    randint(0, 255),
                                    a))


def color_str(string: str,
              fg_color: ColorLike = None,
              bg_color: ColorLike = None) -> str:
    """Returns an ANSI-escaped [#]_ colored representation of `string`.

    Examples:
        >>> print(color_str("It's Christmas!", "#ff0000", "#00ff00"))  # doctest: +SKIP

    Args:
        string: String that will be colored.
        fg_color: Color to be applied to the foreground of the text.
        bg_color: Color to be applied to the background of the text.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/ANSI_escape_code#Colors.
    """
    if fg_color is not None:
        rgba = config.DEFAULT_COLOR_FORMAT.format(fg_color)._rgba
        string = f"\033[38;2;{rgba[0]};{rgba[1]};{rgba[2]}m" + string + "\33[0m"
    if bg_color is not None:
        rgba = config.DEFAULT_COLOR_FORMAT.format(bg_color)._rgba
        string = f"\033[48;2;{rgba[0]};{rgba[1]};{rgba[2]}m" + string + "\33[0m"
    return string


# https://entropymine.com/imageworsener/srgbformula/
def _to_linear_rgb(rgba):
    rgba = list(rgba)
    for i in range(3):
        if rgba[i] <= 255 * 0.04045:
            rgba[i] /= 12.92
        else:
            rgba[i] = 255 * pow(((rgba[i] / 255 + 0.055) / 1.055), 2.4)
    return tuple(rgba)


def _to_srgb(rgba):
    rgba = list(rgba)
    for i in range(3):
        if rgba[i] <= 255 * 0.0031308:
            rgba[i] *= 12.92
        else:
            rgba[i] = 255 * (1.055 * pow(rgba[i] / 255, 1/2.4) - 0.055)
    return tuple(rgba)


import turtle
from colorir import Palette

colors = Palette.load("basic")
colors.add("dreamblue", "#4287f5")
colors.add("swampgreen", "#096b1c")
colors.add("woodbrown", "#8c6941")

turtle.pensize(3)
n_colors = len(colors)  # Get how many colors we currently have
for color in colors:  # Iterate over every color value
    turtle.color(color)
    turtle.forward(800 / n_colors)
    turtle.left(360 / n_colors)
turtle.done()


"""Common color spaces and systems.

This module contains different classes that represent a few of the most common color spaces [#]_.
Colors can be compared, converted to and passed as arguments to different classes.

Examples:
    Create some colors and compare them:

    >>> sRGB(255, 0, 0) == HSL(0, 1 , 0.5)
    True

    Convert an RGB color to CMYK:

    >>> rgb_red = sRGB(255, 0, 0)
    >>> rgb_red.cmyk()
    CMYK(0.0, 1.0, 1.0, 0.0)

References:
    .. [#] Wikipedia at https://en.wikipedia.org/wiki/Color_model.
"""
import abc
import colorsys
import numpy as np
from typing import List
from colormath.color_objects import LabColor, LuvColor, LCHuvColor, sRGBColor, LCHabColor, \
    CMYColor, CMYKColor
from colormath.color_conversions import convert_color

import colorir

__all__ = [
    "ColorBase",
    "ColorTupleBase",
    "ColorPolarBase",
    "sRGB",
    "HSV",
    "HSL",
    "CMY",
    "CMYK",
    "CIELuv",
    "CIELab",
    "HCLuv",
    "HCLab",
    "Hex",
    "HexRGB"
]


class ColorBase(metaclass=abc.ABCMeta):
    """Base class from which all color classes inherit.

    Notes:
        This class is abstract and should not be instantiated.
    """
    _rgba: np.ndarray
    _format_params: List[str]

    # Factory method to be called when reading the palette files or reconstructing colors
    @classmethod
    @abc.abstractmethod
    def _from_rgba(cls, rgba, **kwargs):
        pass

    def __eq__(self, other):
        return all(self._rgba == other._rgba) if isinstance(other, ColorBase) else False

    def __hash__(self):
        return hash(tuple(self._rgba))

    def get_format(self) -> "colorir.color_format.ColorFormat":
        """Returns a :class:`~colorir.color_format.ColorFormat` representing the format of this
        color object."""
        format_ = {param: getattr(self, param) for param in self._format_params}
        return colorir.color_format.ColorFormat(self.__class__, **format_)

    def grayscale(self):
        """Converts this color to a grayscale representation in the same format using CIELuv
        lightness component."""
        return self.get_format().format(CIELuv(self.cieluv().l, 0, 0))

    def hex(self, **kwargs) -> "Hex":
        """Converts the current color to a hexadecimal representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`Hex` constructor.
        """
        return Hex._from_rgba(self._rgba, **kwargs)

    def rgb(self, **kwargs) -> "sRGB":
        """Converts the current color to an sRGB representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`sRGB` constructor.
        """
        return sRGB._from_rgba(self._rgba, **kwargs)

    def hsl(self, **kwargs) -> "HSL":
        """Converts the current color to an HSL representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`HSL` constructor.
        """
        return HSL._from_rgba(self._rgba, **kwargs)

    def hsv(self, **kwargs) -> "HSV":
        """Converts the current color to an HSV representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`HSV` constructor.
        """
        return HSV._from_rgba(self._rgba, **kwargs)

    def cmyk(self, **kwargs) -> "CMYK":
        """Converts the current color to a CMYK representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`CMYK` constructor.
        """
        return CMYK._from_rgba(self._rgba, **kwargs)

    def cmy(self, **kwargs) -> "CMY":
        """Converts the current color to a CMY representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`CMY` constructor.
        """
        return CMY._from_rgba(self._rgba, **kwargs)

    def cieluv(self, **kwargs) -> "CIELuv":
        """Converts the current color to a CIELuv representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`CIELuv` constructor.
        """
        return CIELuv._from_rgba(self._rgba, **kwargs)

    def cielab(self, **kwargs) -> "CIELab":
        """Converts the current color to a CIELab representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`CIELab` constructor.
        """
        return CIELab._from_rgba(self._rgba, **kwargs)

    def hcluv(self, **kwargs) -> "HCLuv":
        """Converts the current color to a HCLuv representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`HCLuv` constructor.
        """
        return HCLuv._from_rgba(self._rgba, **kwargs)

    def hclab(self, **kwargs) -> "HCLab":
        """Converts the current color to a HCLab representation.

        Args:
            **kwargs: Keyword arguments wrapped in this function will be passed on to the
                :class:`HCLab` constructor.
        """
        return HCLab._from_rgba(self._rgba, **kwargs)


class ColorTupleBase(ColorBase, tuple, metaclass=abc.ABCMeta):
    """Base class from which all color classes that are represented by tuples inherit.

    Notes:
        This class is abstract and should not be instantiated.
    """
    @abc.abstractmethod
    def __new__(cls, specs, a, rgba, include_a, round_to):
        specs = list(specs)
        if round_to == 0:
            specs = [round(val) for val in specs]
            a = round(a)
        elif round_to > 0:
            specs = [round(val, round_to) for val in specs]
            a = round(a, round_to)
        if include_a:
            specs.append(a)
        obj = tuple.__new__(cls, specs)
        obj.a = a
        obj.include_a = include_a
        obj.round_to = round_to
        obj._rgba = np.rint(rgba).astype(int)
        obj._format_params = ["include_a", "round_to"]

        return obj

    def __repr__(self):
        return f"{self.__class__.__name__}{tuple.__repr__(self)}"

    # It would be very dangerous to change str conversion as the target framework could call it
    # expecting (255, 0, 0) and get sRGB(255, 0, 0)
    def __str__(self):
        return tuple.__repr__(self)

    def __eq__(self, other):
        if isinstance(other, ColorBase):
            return ColorBase.__eq__(self, other)
        return tuple.__eq__(self, other)

    def __hash__(self):
        return ColorBase.__hash__(self)


class ColorPolarBase(ColorTupleBase, metaclass=abc.ABCMeta):
    """Base class from which all color classes that are represented by tuples and have a hue
     component in polar coordinates inherit.

    Notes:
        This class is abstract and should not be instantiated.
    """
    h: float
    max_h: float


class sRGB(ColorTupleBase):
    """Represents a color in the RGB color space [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/SRGB.

    Args:
        r: Red component of the color.
        g: Green component of the color.
        b: Blue component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max_rgba` parameter.
        max_rgb: What is the maximum value for the `r`, `g` and `b` components. Some common
            values for this parameter would be 255 or 1.
        max_a : What is the maximum value for the `a` component. Some common
            values for this parameter would be 100 or 1.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`sRGB(255, 255, 0,
            255)` instead of :code:`sRGB(255, 255, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. -1
            means that the components won't be rounded at all.
    """
    def __new__(cls,
                r: float,
                g: float,
                b: float,
                a: float = None,
                max_rgb=255,
                max_a=1,
                include_a=False,
                round_to=0):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not all(0 <= spec <= max_rgb for spec in (r, g, b)):
            raise ValueError("'r', 'g' and 'b' must be greater than 0 and smaller than 'max_rgb'")

        rgba = np.array((r, g, b, a), dtype=float) * 255
        rgba[:3] /= max_rgb
        rgba[-1] /= max_a

        obj = super().__new__(
            cls,
            (r, g, b),
            a,
            rgba,
            include_a=include_a,
            round_to=round_to
        )
        obj.r, obj.g, obj.b = obj[:3]
        obj.max_rgb = max_rgb
        obj.max_a = max_a
        obj._format_params += ["max_rgb", "max_a"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_rgb=255, max_a=1, include_a=False, round_to=0):
        rgba_ = np.array(rgba) / 255 * max_rgb

        obj = super().__new__(cls,
                              rgba_[:3],
                              rgba_[-1],
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.r, obj.g, obj.b = obj[:3]
        obj.max_rgb = max_rgb
        obj.max_a = max_a
        return obj


class HSL(ColorPolarBase):
    """Represents a color in the HSL color space [#]_.

    .. [#] Wikipedia at https://en.wikipedia.org/wiki/HSL_and_HSV.

    Args:
        h: HUE component of the color.
        s: Saturation component of the color.
        l: Lightness component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max_sla` parameter.
        max_h:  What is the maximum value for the `h` component. Some common values for this
            parameter would be 360 or 1.
        max_sla:  What is the maximum value for the `s`, `l` and `a` components. Some common values
            for this parameter would be 1 or 100.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`HSL(360, 1, 0,
            1)` instead of :code:`HSL(360, 1, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. -1
            means that the components won't be rounded at all.
    """
    def __new__(cls,
                h: float,
                s: float,
                l: float,
                a: float = None,
                max_h=360,
                max_sla=1,
                include_a=False,
                round_to=-1):
        if a is None:
            a = max_sla
        if not all(0 <= spec <= max_sla for spec in (s, l, a)):
            raise ValueError("'s', 'l' and 'a' must be greater than 0 and smaller than 'max_sla'")

        rgba = colorsys.hls_to_rgb(h % max_h / max_h, l / max_sla, s / max_sla) + (a / max_sla,)

        obj = super().__new__(cls,
                              (h, s, l),
                              a,
                              np.array(rgba) * 255,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.s, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_sla = max_sla
        obj._format_params += ["max_h", "max_sla"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_h=360, max_sla=1, include_a=False, round_to=-1):
        hls = colorsys.rgb_to_hls(*np.array(rgba[:-1]) / 255)
        hsl = (hls[0] * max_h, hls[2] * max_sla, hls[1] * max_sla)

        obj = super().__new__(cls,
                              hsl,
                              rgba[-1] / 255 * max_sla,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.s, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_sla = max_sla
        obj._format_params += ["max_h", "max_sla"]

        return obj


class HSV(ColorPolarBase):
    """Represents a color in the HSV color space [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/HSL_and_HSV.

    Args:
        h: HUE component of the color.
        s: Saturation component of the color.
        v: Value component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max_sva` parameter.
        max_h:  What is the maximum value for the `h` component. Some common values for this
            parameter would be 360 or 1.
        max_sva:  What is the maximum value for the `s`, `v` and `a` components. Some common values
            for this parameter would be 1 or 100.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`HSV(360, 1, 0,
            1)` instead of :code:`HSV(360, 1, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. -1
            means that the components won't be rounded at all.
    """
    def __new__(cls,
                h: float,
                s: float,
                v: float,
                a: float = None,
                max_h=360,
                max_sva=1,
                include_a=False,
                round_to=-1):
        if a is None:
            a = max_sva
        if not all(0 <= spec <= max_sva for spec in (s, v, a)):
            raise ValueError("'s', 'v' and 'a' must be greater than 0 and smaller than 'max_sva'")

        rgba = colorsys.hsv_to_rgb(h % max_h / max_h, s / max_sva, v / max_sva) + (a / max_sva,)

        obj = super().__new__(cls,
                              (h, s, v),
                              a,
                              np.array(rgba) * 255,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.s, obj.v = obj[:3]
        obj.max_h = max_h
        obj.max_sva = max_sva
        obj._format_params += ["max_h", "max_sva"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_h=360, max_sva=1, include_a=False, round_to=-1):
        hsv = colorsys.rgb_to_hsv(*np.array(rgba[:-1]) / 255)
        hsv = (hsv[0] * max_h, hsv[1] * max_sva, hsv[2] * max_sva)

        obj = super().__new__(cls,
                              hsv,
                              rgba[-1] / 255 * max_sva,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.s, obj.v = obj[:3]
        obj.max_h = max_h
        obj.max_sva = max_sva
        obj._format_params += ["max_h", "max_sva"]

        return obj


class CMYK(ColorTupleBase):
    """Represents a color in the CMYK color space [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/CMYK_color_model.

    Args:
        c: Cyan component of the color.
        m: Magenta component of the color.
        y: Yellow component of the color.
        k: Key (black) component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max` parameter.
        max_cmyka: What is the maximum value for the `c`, `m`, `y`, `k` and `a` components. Some
            common values for this parameter would be 1 or 100.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`CMYK(1, 1, 0,
            1)` instead of :code:`CMYK(1, 1, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. The default,
            -1, means that the components won't be rounded at all.
    """
    def __new__(cls, c: float, m: float, y: float, k: float, a: float = None, max_cmyka=1,
                include_a=False,
                round_to=-1):
        if a is None:
            a = max_cmyka
        if not all(0 <= spec <= max_cmyka for spec in (c, m, y, k, a)):
            raise ValueError("'c', 'm', 'y', 'k', and 'a' must be greater than 0 and smaller than "
                             "'max_cmyka'")

        rgba = convert_color(
            CMYKColor(*(np.array((c, m, y, k)) / max_cmyka)),
            sRGBColor
        ).get_value_tuple() + (a / max_cmyka,)

        obj = super().__new__(cls,
                              (c, m, y, k),
                              a,
                              np.array(rgba) * 255,
                              include_a=include_a,
                              round_to=round_to)
        obj.c, obj.m, obj.y, obj.k = obj[:4]
        obj.max_cmyka = max_cmyka
        obj._format_params.append("max_cmyka")

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_cmyka=1, include_a=False, round_to=-1):
        cmyk = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            CMYKColor
        ).get_value_tuple()
        cmyk = np.array(cmyk) * max_cmyka

        obj = super().__new__(cls,
                              cmyk,
                              rgba[-1] / 255 * max_cmyka,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.c, obj.m, obj.y, obj.k = obj[:4]
        obj.max_cmyka = max_cmyka
        obj._format_params.append("max_cmyka")

        return obj


class CMY(ColorTupleBase):
    """Represents a color in the CMY color space [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/CMY_color_model.


    Args:
        c: Cyan component of the color.
        m: Magenta component of the color.
        y: Yellow component of the color.
        a: Opacity component of the color. Defaults to ``None``, which means it will be the same
            as the `max` parameter.
        max_cmya: What is the maximum value for the `c`, `m`, `y` and `a` components. Some common
            values for this parameter would be 1 or 100.
        include_a: Whether to include the opacity parameter `a` in the constructed tuple.
            Setting it to ``True`` may result in an object such as :code:`CMY(1, 1, 0,
            1)` instead of :code:`CMY(1, 1, 0)`, for exemple.
        round_to: Rounds the value of each color component to this many decimal places. Setting
            this parameter to 0 ensures that the components will be of type `int`. The default,
            -1, means that the components won't be rounded at all.
    """
    def __new__(cls,
                c: float,
                m: float,
                y: float,
                a: float = None,
                max_cmya=1,
                include_a=False,
                round_to=-1):
        if a is None:
            a = max_cmya
        if not all(0 <= spec <= max_cmya for spec in (c, m, y, a)):
            raise ValueError("'c', 'm', 'y', and 'a' must be greater than 0 and smaller than "
                             "'max_cmya'")

        rgba = convert_color(
            CMYColor(*(np.array((c, m, y)) / max_cmya)),
            sRGBColor
        ).get_value_tuple() + (a / max_cmya,)

        obj = super().__new__(
            cls,
            (c, m, y),
            a,
            np.array(rgba) * 255,
            include_a=include_a,
            round_to=round_to)
        obj.c, obj.m, obj.y = obj[:3]
        obj.max_cmya = max_cmya
        obj._format_params.append("max_cmya")

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_cmya=1, include_a=False, round_to=-1):
        cmy = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            CMYColor
        ).get_value_tuple()
        cmy = np.array(cmy) * max_cmya

        obj = super().__new__(cls,
                              cmy,
                              rgba[-1] / 255 * max_cmya,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.c, obj.m, obj.y = obj[:3]
        obj.max_cmya = max_cmya
        obj._format_params.append("max_cmya")

        return obj


# TODO doc
class CIELuv(ColorTupleBase):
    def __new__(cls, l, u, v, a=None, max_a=1, include_a=False, round_to=-1):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not 0 <= l <= 100:
            raise ValueError("'l' must be greater than 0 and smaller than 100")

        rgb = convert_color(LuvColor(l, u, v, illuminant="d65"), sRGBColor)
        rgba = (rgb.clamped_rgb_r * 255,
                rgb.clamped_rgb_g * 255,
                rgb.clamped_rgb_b * 255,
                a / max_a * 255)
        obj = super().__new__(cls, (l, u, v), a, rgba, include_a=include_a, round_to=round_to)
        obj.l, obj.u, obj.v = obj[:3]
        obj.max_a = max_a
        obj._format_params.append("max_a")

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_a=1, include_a=False, round_to=-1):
        luv = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            LuvColor,
            target_illuminant="d65"
        ).get_value_tuple()
        obj = super().__new__(cls,
                              luv,
                              rgba[-1] / 255 * max_a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.l, obj.u, obj.v = obj[:3]
        obj.max_a = max_a
        obj._format_params.append("max_a")

        return obj


# TODO doc
class CIELab(ColorTupleBase):
    def __new__(cls, l, a_, b, a=None, max_a=1, include_a=False, round_to=-1):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not 0 <= l <= 100:
            raise ValueError("'l' must be greater than 0 and smaller than 100")

        rgb = convert_color(LabColor(l, a_, b, illuminant="d65"), sRGBColor)
        rgba = (rgb.clamped_rgb_r * 255,
                rgb.clamped_rgb_g * 255,
                rgb.clamped_rgb_b * 255,
                a / max_a * 255)
        obj = super().__new__(cls, (l, a_, b), a, rgba, include_a=include_a, round_to=round_to)
        obj.l, obj.a_, obj.b = l, a_, b
        obj.max_a = max_a
        obj._format_params.append("max_a")

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_a=1, include_a=False, round_to=-1):
        lab = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            LabColor,
            target_illuminant="d65"
        ).get_value_tuple()

        obj = super().__new__(cls,
                              lab,
                              rgba[-1] / 255 * max_a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.l, obj.a_, obj.b = obj[:3]
        obj.max_a = max_a
        obj._format_params.append("max_a")

        return obj


# TODO doc
class HCLuv(ColorPolarBase):
    def __new__(cls, h, c, l, a=None, max_h=360, max_a=1, include_a=False, round_to=-1):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not 0 <= l <= 100:
            raise ValueError("'l' must be greater than 0 and smaller than 100")

        rgb = convert_color(
            LCHuvColor(l, c, h / max_h * 360 % 360, illuminant="d65"),
            sRGBColor
        )
        rgba = (rgb.clamped_rgb_r * 255,
                rgb.clamped_rgb_g * 255,
                rgb.clamped_rgb_b * 255,
                a / max_a * 255)
        obj = super().__new__(cls,
                              (h, c, l),
                              a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.c, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_sla = max_a
        obj._format_params += ["max_h", "max_a"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_h=360, max_a=1, include_a=False, round_to=-1):
        hcl = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            LCHuvColor,
            target_illuminant="d65"
        ).get_value_tuple()[::-1]
        hcl = np.array(hcl)
        hcl[0] *= max_h / 360
        obj = super().__new__(
            cls,
            hcl,
            rgba[-1] / 255 * max_a,
            rgba,
            include_a=include_a,
            round_to=round_to
        )
        obj.h, obj.c, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_a = max_a
        obj._format_params += ["max_h", "max_a"]

        return obj


# TODO doc
class HCLab(ColorPolarBase):
    def __new__(cls, h, c, l, a=None, max_h=360, max_a=1, include_a=False, round_to=-1):
        if a is None:
            a = max_a
        elif not 0 <= a <= max_a:
            raise ValueError("'a' must be greater than 0 and smaller than 'max_a'")
        if not 0 <= l <= 100:
            raise ValueError("'l' must be greater than 0 and smaller than 100")

        rgb = convert_color(
            LCHabColor(l, c, h / max_h * 360 % 360, illuminant="d65"),
            sRGBColor
        )
        rgba = (rgb.clamped_rgb_r * 255,
                rgb.clamped_rgb_g * 255,
                rgb.clamped_rgb_b * 255,
                a / max_a * 255)

        obj = super().__new__(cls,
                              (h, c, l),
                              a,
                              rgba,
                              include_a=include_a,
                              round_to=round_to)
        obj.h, obj.c, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_sla = max_a
        obj._format_params += ["max_h", "max_a"]

        return obj

    @classmethod
    def _from_rgba(cls, rgba, max_h=360, max_a=1, include_a=False, round_to=-1):
        hcl = convert_color(
            sRGBColor(*rgba[:3], is_upscaled=True),
            LCHabColor,
            target_illuminant="d65"
        ).get_value_tuple()[::-1]
        hcl = np.array(hcl)
        hcl[0] *= max_h / 360
        obj = super().__new__(
            cls,
            hcl,
            rgba[-1] / 255 * max_a,
            rgba,
            include_a=include_a,
            round_to=round_to
        )
        obj.h, obj.c, obj.l = obj[:3]
        obj.max_h = max_h
        obj.max_a = max_a
        obj._format_params += ["max_h", "max_a"]

        return obj


class Hex(ColorBase, str):
    """Represents a color in the RGB color space [#]_ as a hexadecimal string.

    Is mostly used for representing colors in web applications [#]_.

    References:
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/SRGB.
        .. [#] Wikipedia at https://en.wikipedia.org/wiki/Web_colors

    Args:
        hex_str: Hexadecimal string from which the :class:`Hex` instance will be built.
            May or may not include a "#" character in its beginning.
        uppercase: Whether the color will be represented in uppercase or lowercase.
        include_a: Whether to include the opacity parameter `a` in the constructed string.
            Setting it to ``True`` may result in an object such as :code:`Hex('#ffffff00')`
            instead of :code:`Hex('#ffff00')`, for exemple.
        tail_a: Whether the alpha component is present in the tail or head of the hex string. Used
            only if `hex_str` includes an alpha component or `include_a` is ``True``.

    Examples:
        >>> Hex("#ff0000")
        Hex('#ff0000')

        >>> Hex("#FF0000", include_hash=False)
        Hex('ff0000')

        >>> Hex("ff0000", include_a=True, tail_a=True)
        Hex('#ff0000ff')
    """
    def __new__(cls,
                hex_str: str,
                uppercase=False,
                include_hash=True,
                include_a=False,
                tail_a=False):
        hex_str = hex_str.lstrip("#")
        if not 6 <= len(hex_str) <= 9:
            raise ValueError("'hex_str' length must be between 6 and 9 (inclusive)")
        if len(hex_str) < 8:
            i = 0
            a = 255
        elif tail_a:
            i = 0
            a = int(hex_str[-2:], 16)
        else:
            i = 2
            a = int(hex_str[:2], 16)
        rgb = (int(hex_str[i:i + 2], 16),
               int(hex_str[i + 2:i + 4], 16),
               int(hex_str[i + 4:i + 6], 16))
        # Rather than modifying the stored string to suit the format specifications we just
        # delegate that to the _from_rgba method
        return cls._from_rgba(rgb + (a,),
                              uppercase=uppercase,
                              include_hash=include_hash,
                              include_a=include_a,
                              tail_a=tail_a)

    @classmethod
    def _from_rgba(cls, rgba, uppercase=False, include_hash=True, include_a=False, tail_a=False):
        rgba = tuple(rgba)
        hex_str = '%02x%02x%02x' % rgba[:3]

        if include_a:
            a = '%02x' % rgba[-1]
            if not tail_a:
                hex_str = a + hex_str
            else:
                hex_str += a
        if uppercase:
            hex_str = hex_str.upper()
        if include_hash:
            hex_str = '#' + hex_str

        obj = str.__new__(cls, hex_str)
        obj.uppercase = uppercase
        obj.include_hash = include_hash
        obj.include_a = include_a
        obj.tail_a = tail_a
        obj._rgba = np.rint(rgba).astype(int)
        obj._format_params = ["uppercase", "include_hash", "include_a", "tail_a"]
        return obj

    def __repr__(self):
        return f"{self.__class__.__name__}({str.__repr__(self)})"

    # It would be very dangerous to change str conversion as the target framework could call it
    # expecting #ff0000 and get Hex('#ff0000')
    def __str__(self):
        return str.__str__(self)

    def __eq__(self, other):
        if isinstance(other, ColorBase):
            return ColorBase.__eq__(self, other)
        return str.__eq__(self, other)

    def __hash__(self):
        return ColorBase.__hash__(self)


# Alias
HexRGB = Hex